import os

# Transformers is a dependency of DeepSeek, and it logs a lot of warnings that are not useful
os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "1"

from logging import getLogger

from google.genai.types import HttpOptions
from llama_index.llms.anthropic import Anthropic
from llama_index.llms.deepseek import DeepSeek
from llama_index.llms.google_genai import GoogleGenAI
from llama_index.llms.groq import Groq
from llama_index.llms.mistralai import MistralAI
from llama_index.llms.openai import OpenAI

from ..common.settings import settings
from .custom.llamaindex.cohere import Cohere

logger = getLogger(__name__)

async def get_llamaindex_model(url: str, type: str, model: str, **kwargs):
    if type == 'anthropic':
        return Anthropic(
            model=model,
            api_key=settings.auth.token,
            base_url=url,
            default_headers=settings.auth.get_headers(),
            **kwargs
        )
    elif type == 'xai':
        return Groq(
            model=model,
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'gemini':
        return GoogleGenAI(
            api_key=settings.auth.token,
            model=model,
            api_base=f"{url}/v1",
            http_options=HttpOptions(
                base_url=url,
                headers=settings.auth.get_headers(),
            ),
            **kwargs
        )
    elif type == 'cohere':
        return Cohere(
            model=model,
            api_key=settings.auth.token,
            api_base=url,
            **kwargs
        )
    elif type == 'deepseek':
        return DeepSeek(
            model=model,
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'mistral':
        return MistralAI(
            model=model,
            api_key=settings.auth.token,
            endpoint=url,
            **kwargs
        )
    else:
        if type != "openai":
            logger.warning(f"Model {model} is not supported by LlamaIndex, defaulting to OpenAI")

        return OpenAI(
            model=model,
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )


from dataclasses import dataclass, field
from typing import Optional, Union

from ..client import Client
from ..client import client as default_client


@dataclass
class OauthTokenData:
    body: dict[str, Union[str, Optional[str]]] = field(default_factory=dict)
    headers: dict[str, str] = field(default_factory=dict)
    authenticated: Optional[bool] = False

@dataclass
class OauthTokenResponse:
    access_token: str
    refresh_token: str
    expires_in: int
    token_type: str

@dataclass
class OauthTokenError:
    error: str

async def oauth_token(
    options: OauthTokenData,
    client: Optional[Client] = None,
    throw_on_error: bool = False
) -> Union[OauthTokenResponse, OauthTokenError]:
    """
    Get a new OAuth token.

    Args:
        options: The OAuth token request options
        client: Optional client instance to use for the request
        throw_on_error: Whether to throw an exception on error

    Returns:
        The OAuth token response or error
    """
    response = await (client or default_client).get_async_httpx_client().post(
        url="/oauth/token",
        json=options.body or {},
        headers=options.headers or {},
    )
    if response.status_code >= 400:
        if throw_on_error:
            raise Exception(f"Failed to get OAuth token: {response.text}")
        return OauthTokenError(error=response.text)
    return OauthTokenResponse(**response.json())


"""
This module provides utilities for creating and managing OpenTelemetry spans within Blaxel.
It includes classes for adding default attributes to spans and managing span creation.
"""

from typing import Any, ContextManager, Dict, Optional, TypeVar

from opentelemetry import trace
from opentelemetry.sdk.trace import Span as SdkSpan
from opentelemetry.sdk.trace.export import SpanProcessor
from opentelemetry.trace import Span, Tracer

from ..common.settings import settings

T = TypeVar('T')


class DefaultAttributesSpanProcessor(SpanProcessor):
    """A span processor that adds default attributes to spans when they are created."""

    def __init__(self, default_attributes: Dict[str, str]):
        self.default_attributes = default_attributes

    def on_start(self, span: SdkSpan, parent_context = None) -> None:
        """Add default attributes to the span when it starts."""
        for key, value in self.default_attributes.items():
            span.set_attribute(key, value)

    def on_end(self, span: SdkSpan) -> None:
        """Called when a span ends."""
        pass

    def shutdown(self) -> None:
        """Shuts down the span processor."""
        pass

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        """Forces the span processor to flush any queued spans."""
        return True


class SpanManager:
    """Manages the creation and lifecycle of spans."""

    def __init__(self, name: str):
        self.tracer: Tracer = trace.get_tracer(name)

    def create_active_span(self, name: str, attributes: Dict[str, Any], parent: Optional[Span] = None) -> ContextManager[Span]:
        """
        Creates an active span and executes the provided function within its context.

        Args:
            name: The name of the span
            attributes: Attributes to add to the span
            parent: Optional parent span

        Returns:
            Context manager that yields the span
        """
        # Add default attributes
        attributes["blaxel.environment"] = settings.env
        attributes["workload.id"] = settings.name
        attributes["workload.type"] = f"{settings.type}s"
        attributes["workspace"] = settings.workspace
        context = None
        if parent:
            context = trace.set_span_in_context(parent)
        return self.tracer.start_as_current_span(name, attributes=attributes, context=context)

    def create_span(self, name: str, attributes: Dict[str, Any], parent: Optional[Span] = None) -> Span:
        """
        Creates a new span without making it active.

        Args:
            name: The name of the span
            attributes: Attributes to add to the span
            parent: Optional parent span

        Returns:
            The created span
        """
        # Add default attributes
        full_attributes = {
            **attributes,
            "blaxel.environment": settings.env,
            "workload.id": settings.name,
            "workload.type": f"{settings.type}s",
            "workspace": settings.workspace
        }

        context = None
        if parent:
            context = trace.set_span_in_context(parent)
        return self.tracer.start_span(name, attributes=full_attributes, context=context)



import asyncio
import os
from logging import getLogger

from dotenv import load_dotenv

load_dotenv()

logger = getLogger(__name__)

async def test_blaxel_config():
    from blaxel.client import client
    from blaxel.client.api.models import list_models

    models = await list_models.asyncio(client=client)
    logger.info([model.metadata.name for model in models])

async def test_blaxel_client_credentials():
    os.environ["BL_WORKSPACE"] = os.getenv("BL_WORKSPACE") or input("Enter workspace: ")
    os.environ["BL_CLIENT_CREDENTIALS"] = os.getenv("BL_CLIENT_CREDENTIALS") or input("Enter client credentials: ")
    os.environ["BL_ENV"] = "dev"

    from blaxel.client import client
    from blaxel.client.api.models import list_models

    models = await list_models.asyncio(client=client)
    logger.info([model.metadata.name for model in models])

async def test_blaxel_api_key():
    os.environ["BL_WORKSPACE"] = os.getenv("BL_WORKSPACE") or input("Enter workspace: ")
    os.environ["BL_API_KEY"] = os.getenv("BL_API_KEY") or input("Enter API key: ")
    os.environ["BL_ENV"] = "dev"

    from blaxel.client import client
    from blaxel.client.api.models import list_models

    models = await list_models.asyncio(client=client)
    logger.info([model.metadata.name for model in models])


if __name__ == "__main__":
    # You can only execute one test at a time, cause of autoload
    asyncio.run(test_blaxel_config())
    # asyncio.run(test_blaxel_client_credentials())
    # asyncio.run(test_blaxel_api_key())


import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from blaxel.models import bl_model

logger = getLogger(__name__)


MODEL = "gpt-4o-mini"
# MODEL = "claude-3-5-sonnet"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus"
# MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"

async def test_model_langchain():
    model = await bl_model(MODEL).to_langchain()
    result = await model.ainvoke("Hello, world!")
    logger.info(result)

async def test_model_llamaindex():
    model = await bl_model(MODEL).to_llamaindex()
    result = await model.acomplete("Hello, world!")
    logger.info(result)

async def test_model_crewai():
    # not working: cohere
    model = await bl_model(MODEL).to_crewai()
    result = model.call([{"role": "user", "content": "Hello, world!"}])
    logger.info(result)

async def main():
    await test_model_langchain()
    await test_model_llamaindex()
    await test_model_crewai()

if __name__ == "__main__":
    asyncio.run(main())


from __future__ import annotations

import asyncio
import json
import logging
import uuid
import warnings
from operator import itemgetter
from typing import (Any, AsyncIterator, Callable, Dict, Iterator, List,
                    Mapping, Optional, Sequence, Tuple, Type, Union, cast)

import httpx
import requests
from langchain_core.callbacks.manager import (AsyncCallbackManagerForLLMRun,
                                              CallbackManagerForLLMRun)
from langchain_core.language_models import LanguageModelInput
from langchain_core.language_models.chat_models import (BaseChatModel,
                                                        LangSmithParams)
from langchain_core.messages import (AIMessage, AIMessageChunk, BaseMessage,
                                     FunctionMessage, HumanMessage,
                                     SystemMessage, ToolMessage)
from langchain_core.messages.ai import UsageMetadata
from langchain_core.messages.tool import (invalid_tool_call, tool_call,
                                          tool_call_chunk)
from langchain_core.output_parsers.openai_tools import (
    JsonOutputKeyToolsParser, PydanticToolsParser, parse_tool_calls)
from langchain_core.outputs import (ChatGeneration, ChatGenerationChunk,
                                    ChatResult)
from langchain_core.runnables import Runnable, RunnablePassthrough
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from PIL import Image
from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from tenacity import (before_sleep_log, retry, retry_if_exception_type,
                      stop_after_attempt, wait_exponential)
from typing_extensions import Self, is_typeddict

WARNED_STRUCTURED_OUTPUT_JSON_MODE = False

logger = logging.getLogger(__name__)

# Type definitions
_FunctionDeclarationType = Union[
    Dict[str, Any],
    Callable[..., Any],
]

_ToolChoiceType = Union[str, bool, Dict[str, Any]]
_ToolConfigDict = Dict[str, Any]
_ToolDict = Dict[str, Any]
SafetySettingDict = Dict[str, str]
OutputParserLike = Union[PydanticToolsParser, JsonOutputKeyToolsParser]

# Data classes
class Part(BaseModel):
    text: Optional[str] = None
    inline_data: Optional[Dict[str, Any]] = None
    file_data: Optional[Dict[str, Any]] = None
    function_call: Optional[Dict[str, Any]] = None
    function_response: Optional[Dict[str, Any]] = None

class Content(BaseModel):
    role: Optional[str] = None
    parts: List[Part]

class Blob(BaseModel):
    data: str
    mime_type: str

class FileData(BaseModel):
    file_uri: str
    mime_type: str

class VideoMetadata(BaseModel):
    duration: Optional[str] = None
    start_offset: Optional[str] = None
    end_offset: Optional[str] = None

class FunctionCall(BaseModel):
    name: str
    args: Dict[str, Any]

class FunctionResponse(BaseModel):
    name: str
    response: Dict[str, Any]

class SafetySetting(BaseModel):
    category: str
    threshold: str

class ToolConfig(BaseModel):
    function_calling_config: Dict[str, Any]

class GenerationConfig(BaseModel):
    candidate_count: Optional[int] = None
    temperature: Optional[float] = None
    stop_sequences: Optional[List[str]] = None
    max_output_tokens: Optional[int] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    response_modalities: Optional[List[str]] = None

class GoogleTool(BaseModel):
    name: str
    description: str
    parameters: Dict[str, Any]

class ImageBytesLoader:
    def load_part(self, image_url: str) -> Part:
        """Load an image from a URL and convert it to a Part."""
        import io

        response = requests.get(image_url)
        response.raise_for_status()

        # Convert to JPEG format
        img = Image.open(io.BytesIO(response.content))
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format='JPEG')
        img_byte_arr = img_byte_arr.getvalue()

        return Part(
            inline_data={
                "mime_type": "image/jpeg",
                "data": img_byte_arr.decode('utf-8')
            }
        )

class ChatGoogleGenerativeAIError(Exception):
    """
    Custom exception class for errors associated with the `Google GenAI` API.

    This exception is raised when there are specific issues related to the
    Google genai API usage in the ChatGoogleGenerativeAI class, such as unsupported
    message types or roles.
    """
    pass


def _create_retry_decorator() -> Callable[[Any], Any]:
    """
    Creates and returns a preconfigured tenacity retry decorator.

    The retry decorator is configured to handle specific HTTP exceptions
    such as rate limits and service unavailability. It uses an exponential
    backoff strategy for retries.

    Returns:
        Callable[[Any], Any]: A retry decorator configured for handling specific
        HTTP exceptions.
    """
    multiplier = 2
    min_seconds = 1
    max_seconds = 60
    max_retries = 2

    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=multiplier, min=min_seconds, max=max_seconds),
        retry=(
            retry_if_exception_type(httpx.HTTPStatusError)
            | retry_if_exception_type(httpx.ConnectError)
            | retry_if_exception_type(httpx.ReadTimeout)
        ),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def _chat_with_retry(generation_method: Callable, **kwargs: Any) -> Any:
    """
    Executes a chat generation method with retry logic using tenacity.

    This function is a wrapper that applies a retry mechanism to a provided
    chat generation function. It is useful for handling intermittent issues
    like network errors or temporary service unavailability.

    Args:
        generation_method (Callable): The chat generation method to be executed.
        **kwargs (Any): Additional keyword arguments to pass to the generation method.

    Returns:
        Any: The result from the chat generation method.
    """
    retry_decorator = _create_retry_decorator()

    @retry_decorator
    def _chat_with_retry(**kwargs: Any) -> Any:
        try:
            # Extract request parameters and other kwargs
            request = kwargs.pop('request', {})
            metadata = kwargs.pop('metadata', None)

            # Unpack request parameters into kwargs
            kwargs.update(request)

            return generation_method(**kwargs)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 429:  # Rate limit
                error_msg = (
                    "Rate limit exceeded. Please try again later or use a different API key."
                )
                raise ValueError(error_msg)
            elif e.response.status_code == 403:  # Forbidden
                error_msg = (
                    "Access forbidden. Please check your API key and permissions."
                )
                raise ValueError(error_msg)
            else:
                raise ChatGoogleGenerativeAIError(
                    f"HTTP error occurred: {e.response.text}"
                ) from e
        except Exception as e:
            raise e

    return _chat_with_retry(**kwargs)


async def _achat_with_retry(generation_method: Callable, **kwargs: Any) -> Any:
    """
    Executes a chat generation method with retry logic using tenacity.

    This function is a wrapper that applies a retry mechanism to a provided
    chat generation function. It is useful for handling intermittent issues
    like network errors or temporary service unavailability.

    Args:
        generation_method (Callable): The chat generation method to be executed.
        **kwargs (Any): Additional keyword arguments to pass to the generation method.

    Returns:
        Any: The result from the chat generation method.
    """
    retry_decorator = _create_retry_decorator()

    @retry_decorator
    async def _achat_with_retry(**kwargs: Any) -> Any:
        try:
            # Extract request parameters and other kwargs
            request = kwargs.pop('request', {})
            metadata = kwargs.pop('metadata', None)

            # Unpack request parameters into kwargs
            kwargs.update(request)

            return await generation_method(**kwargs)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 429:  # Rate limit
                error_msg = (
                    "Rate limit exceeded. Please try again later or use a different API key."
                )
                raise ValueError(error_msg)
            elif e.response.status_code == 403:  # Forbidden
                error_msg = (
                    "Access forbidden. Please check your API key and permissions."
                )
                raise ValueError(error_msg)
            else:
                raise ChatGoogleGenerativeAIError(
                    f"HTTP error occurred: {e.response.text}"
                ) from e
        except Exception as e:
            raise e

    return await _achat_with_retry(**kwargs)


def _is_openai_parts_format(part: dict) -> bool:
    return "type" in part


def _convert_to_parts(
    raw_content: Union[str, Sequence[Union[str, dict]]],
) -> List[Part]:
    """Converts a list of LangChain messages into a google parts."""
    parts = []
    content = [raw_content] if isinstance(raw_content, str) else raw_content
    image_loader = ImageBytesLoader()
    for part in content:
        if isinstance(part, str):
            parts.append(Part(text=part))
        elif isinstance(part, Mapping):
            # OpenAI Format
            if _is_openai_parts_format(part):
                if part["type"] == "text":
                    parts.append(Part(text=part["text"]))
                elif part["type"] == "image_url":
                    img_url = part["image_url"]
                    if isinstance(img_url, dict):
                        if "url" not in img_url:
                            raise ValueError(
                                f"Unrecognized message image format: {img_url}"
                            )
                        img_url = img_url["url"]
                    parts.append(image_loader.load_part(img_url))
                # Handle media type like LangChain.js
                # https://github.com/langchain-ai/langchainjs/blob/e536593e2585f1dd7b0afc187de4d07cb40689ba/libs/langchain-google-common/src/utils/gemini.ts#L93-L106
                elif part["type"] == "media":
                    if "mime_type" not in part:
                        raise ValueError(f"Missing mime_type in media part: {part}")
                    mime_type = part["mime_type"]
                    media_part = Part()

                    if "data" in part:
                        media_part.inline_data = Blob(
                            data=part["data"], mime_type=mime_type
                        )
                    elif "file_uri" in part:
                        media_part.file_data = FileData(
                            file_uri=part["file_uri"], mime_type=mime_type
                        )
                    else:
                        raise ValueError(
                            f"Media part must have either data or file_uri: {part}"
                        )
                    parts.append(media_part)
                else:
                    raise ValueError(
                        f"Unrecognized message part type: {part['type']}. Only text, "
                        f"image_url, and media types are supported."
                    )
            else:
                # Yolo
                logger.warning(
                    "Unrecognized message part format. Assuming it's a text part."
                )
                parts.append(Part(text=str(part)))
        else:
            # TODO: Maybe some of Google's native stuff
            # would hit this branch.
            raise ChatGoogleGenerativeAIError(
                "Gemini only supports text and inline_data parts."
            )
    return parts


def _convert_tool_message_to_part(
    message: ToolMessage | FunctionMessage, name: Optional[str] = None
) -> Part:
    """Converts a tool or function message to a google part."""
    # Legacy agent stores tool name in message.additional_kwargs instead of message.name
    name = message.name or name or message.additional_kwargs.get("name")
    response: Any
    if not isinstance(message.content, str):
        response = message.content
    else:
        try:
            response = json.loads(message.content)
        except json.JSONDecodeError:
            response = message.content  # leave as str representation
    if not isinstance(response, dict):
        response = {"output": response}
    part = Part(
        function_response=FunctionResponse(
            name=name,
            response=response,
        ).model_dump()
    )
    return part


def _get_ai_message_tool_messages_parts(
    tool_messages: Sequence[ToolMessage], ai_message: AIMessage
) -> list[Part]:
    """
    Finds relevant tool messages for the AI message and converts them to a single
    list of Parts.
    """
    # We are interested only in the tool messages that are part of the AI message
    tool_calls_ids = {tool_call["id"]: tool_call for tool_call in ai_message.tool_calls}
    parts = []
    for i, message in enumerate(tool_messages):
        if not tool_calls_ids:
            break
        if message.tool_call_id in tool_calls_ids:
            tool_call = tool_calls_ids[message.tool_call_id]
            part = _convert_tool_message_to_part(message, name=tool_call.get("name"))
            parts.append(part)
            # remove the id from the dict, so that we do not iterate over it again
            tool_calls_ids.pop(message.tool_call_id)
    return parts


def _parse_chat_history(
    input_messages: Sequence[BaseMessage], convert_system_message_to_human: bool = False
) -> Tuple[Optional[Content], List[Content]]:
    messages: List[Content] = []

    if convert_system_message_to_human:
        warnings.warn("Convert_system_message_to_human will be deprecated!")

    system_instruction: Optional[Content] = None
    messages_without_tool_messages = [
        message for message in input_messages if not isinstance(message, ToolMessage)
    ]
    tool_messages = [
        message for message in input_messages if isinstance(message, ToolMessage)
    ]
    for i, message in enumerate(messages_without_tool_messages):
        if isinstance(message, SystemMessage):
            system_parts = _convert_to_parts(message.content)
            if i == 0:
                system_instruction = Content(parts=system_parts)
            elif system_instruction is not None:
                system_instruction.parts.extend(system_parts)
            else:
                pass
            continue
        elif isinstance(message, AIMessage):
            role = "model"
            if message.tool_calls:
                ai_message_parts = []
                for tool_call in message.tool_calls:
                    function_call = FunctionCall(
                        name=tool_call["name"],
                        args=tool_call["args"],
                    )
                    ai_message_parts.append(Part(function_call=function_call.model_dump()))
                tool_messages_parts = _get_ai_message_tool_messages_parts(
                    tool_messages=tool_messages, ai_message=message
                )
                messages.append(Content(role=role, parts=ai_message_parts))
                messages.append(Content(role="user", parts=tool_messages_parts))
                continue
            elif raw_function_call := message.additional_kwargs.get("function_call"):
                function_call = FunctionCall(
                    name=raw_function_call["name"],
                    args=json.loads(raw_function_call["arguments"]),
                )
                parts = [Part(function_call=function_call.model_dump())]
            else:
                parts = _convert_to_parts(message.content)
        elif isinstance(message, HumanMessage):
            role = "user"
            parts = _convert_to_parts(message.content)
            if i == 1 and convert_system_message_to_human and system_instruction:
                parts = [p for p in system_instruction.parts] + parts
                system_instruction = None
        elif isinstance(message, FunctionMessage):
            role = "user"
            parts = [_convert_tool_message_to_part(message)]
        else:
            raise ValueError(
                f"Unexpected message with type {type(message)} at the position {i}."
            )

        messages.append(Content(role=role, parts=parts))
    return system_instruction, messages


def _parse_response_candidate(
    response_candidate: Dict[str, Any], streaming: bool = False
) -> AIMessage:
    content: Union[None, str, List[Union[str, dict]]] = None
    additional_kwargs = {}
    tool_calls = []
    invalid_tool_calls = []
    tool_call_chunks = []

    for part in response_candidate.get("content", {}).get("parts", []):
        try:
            text: Optional[str] = part.get("text")
            # Remove erroneous newline character if present
            if text is not None:
                text = text.rstrip("\n")
        except AttributeError:
            text = None

        if text is not None:
            if not content:
                content = text
            elif isinstance(content, str) and text:
                content = [content, text]
            elif isinstance(content, list) and text:
                content.append(text)
            elif text:
                raise Exception("Unexpected content type")

        if part.get("inlineData", {}).get("mimeType", "").startswith("image/"):
            image_format = part["inlineData"]["mimeType"][6:]
            message = {
                "type": "image_url",
                "image_url": {
                    "url": image_bytes_to_b64_string(
                        part["inlineData"]["data"], image_format=image_format
                    )
                },
            }

            if not content:
                content = [message]
            elif isinstance(content, str) and message:
                content = [content, message]
            elif isinstance(content, list) and message:
                content.append(message)
            elif message:
                raise Exception("Unexpected content type")

        if part.get("functionCall"):
            function_call = {"name": part["functionCall"]["name"]}
            function_call["arguments"] = json.dumps(part["functionCall"].get("args", {}))
            additional_kwargs["function_call"] = function_call

            if streaming:
                tool_call_chunks.append(
                    tool_call_chunk(
                        name=function_call.get("name"),
                        args=function_call.get("arguments"),
                        id=function_call.get("id", str(uuid.uuid4())),
                        index=function_call.get("index"),  # type: ignore
                    )
                )
            else:
                try:
                    tool_call_dict = parse_tool_calls(
                        [{"function": function_call}],
                        return_id=False,
                    )[0]
                except Exception as e:
                    invalid_tool_calls.append(
                        invalid_tool_call(
                            name=function_call.get("name"),
                            args=function_call.get("arguments"),
                            id=function_call.get("id", str(uuid.uuid4())),
                            error=str(e),
                        )
                    )
                else:
                    tool_calls.append(
                        tool_call(
                            name=tool_call_dict["name"],
                            args=tool_call_dict["args"],
                            id=tool_call_dict.get("id", str(uuid.uuid4())),
                        )
                    )
    if content is None:
        content = ""

    if streaming:
        return AIMessageChunk(
            content=cast(Union[str, List[Union[str, Dict[Any, Any]]]], content),
            additional_kwargs=additional_kwargs,
            tool_call_chunks=tool_call_chunks,
        )

    return AIMessage(
        content=cast(Union[str, List[Union[str, Dict[Any, Any]]]], content),
        additional_kwargs=additional_kwargs,
        tool_calls=tool_calls,
        invalid_tool_calls=invalid_tool_calls,
    )


def _response_to_result(
    response: Dict[str, Any],
    stream: bool = False,
    prev_usage: Optional[UsageMetadata] = None,
) -> ChatResult:
    """Converts a Gemini API response into a LangChain ChatResult."""
    llm_output = {"prompt_feedback": response.get("promptFeedback", {})}

    # previous usage metadata needs to be subtracted because gemini api returns
    # already-accumulated token counts with each chunk
    prev_input_tokens = prev_usage["input_tokens"] if prev_usage else 0
    prev_output_tokens = prev_usage["output_tokens"] if prev_usage else 0
    prev_total_tokens = prev_usage["total_tokens"] if prev_usage else 0

    # Get usage metadata
    try:
        usage_metadata = response.get("usageMetadata", {})
        input_tokens = usage_metadata.get("promptTokenCount", 0)
        output_tokens = usage_metadata.get("candidatesTokenCount", 0)
        total_tokens = usage_metadata.get("totalTokenCount", 0)
        cache_read_tokens = usage_metadata.get("cachedContentTokenCount", 0)
        if input_tokens + output_tokens + cache_read_tokens + total_tokens > 0:
            lc_usage = UsageMetadata(
                input_tokens=input_tokens - prev_input_tokens,
                output_tokens=output_tokens - prev_output_tokens,
                total_tokens=total_tokens - prev_total_tokens,
                input_token_details={"cache_read": cache_read_tokens},
            )
        else:
            lc_usage = None
    except AttributeError:
        lc_usage = None

    generations: List[ChatGeneration] = []

    for candidate in response.get("candidates", []):
        generation_info = {}
        if candidate.get("finishReason"):
            generation_info["finish_reason"] = candidate["finishReason"]
        generation_info["safety_ratings"] = candidate.get("safetyRatings", [])
        message = _parse_response_candidate(candidate, streaming=stream)
        message.usage_metadata = lc_usage
        if stream:
            generations.append(
                ChatGenerationChunk(
                    message=cast(AIMessageChunk, message),
                    generation_info=generation_info,
                )
            )
        else:
            generations.append(
                ChatGeneration(message=message, generation_info=generation_info)
            )
    if not response.get("candidates"):
        # Likely a "prompt feedback" violation (e.g., toxic input)
        # Raising an error would be different than how OpenAI handles it,
        # so we'll just log a warning and continue with an empty message.
        logger.warning(
            "Gemini produced an empty response. Continuing with empty message\n"
            f"Feedback: {response.get('promptFeedback')}"
        )
        if stream:
            generations = [
                ChatGenerationChunk(
                    message=AIMessageChunk(content=""), generation_info={}
                )
            ]
        else:
            generations = [ChatGeneration(message=AIMessage(""), generation_info={})]
    return ChatResult(generations=generations, llm_output=llm_output)


def _is_event_loop_running() -> bool:
    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False


class GeminiRestClient:
    """A REST client for making requests to the Gemini API."""

    def __init__(
        self,
        api_key: str,
        base_url: str,
        headers: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = None,
    ):
        self.api_key = api_key
        self.base_url = base_url.rstrip("/")
        self.headers = headers or {}
        self.timeout = timeout
        self._client = None
        self._async_client = None

    def _get_client(self) -> httpx.Client:
        if self._client is None:
            self._client = httpx.Client(
                base_url=self.base_url,
                headers={
                    "Content-Type": "application/json",
                    **self.headers,
                },
                timeout=self.timeout,
            )
        return self._client

    async def _get_async_client(self) -> httpx.AsyncClient:
        if self._async_client is None:
            self._async_client = httpx.AsyncClient(
                base_url=self.base_url,
                headers={
                    "Content-Type": "application/json",
                    **self.headers,
                },
                timeout=self.timeout,
            )
        return self._async_client

    def _convert_to_dict(self, obj: Any) -> Any:
        """Convert Pydantic models and other objects to dictionaries."""
        if hasattr(obj, 'model_dump'):
            return obj.model_dump()
        elif isinstance(obj, list):
            return [self._convert_to_dict(item) for item in obj]
        elif isinstance(obj, dict):
            return {k: self._convert_to_dict(v) for k, v in obj.items()}
        return obj

    def _prepare_payload(
        self,
        contents: List[Dict[str, Any]],
        generation_config: Optional[Dict[str, Any]] = None,
        safety_settings: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_config: Optional[Dict[str, Any]] = None,
        system_instruction: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Prepare the payload for Gemini API requests."""
        payload = {
            "contents": self._convert_to_dict(contents),
            "generationConfig": self._convert_to_dict(generation_config or {}),
            "safetySettings": self._convert_to_dict(safety_settings or []),
        }
        if tools:
            payload["tools"] = self._convert_to_dict(tools)
        if tool_config:
            payload["toolConfig"] = self._convert_to_dict(tool_config)
        if system_instruction:
            payload["systemInstruction"] = self._convert_to_dict(system_instruction)
        if cached_content:
            payload["cachedContent"] = cached_content
        return payload

    def generate_content(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Optional[Dict[str, Any]] = None,
        safety_settings: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_config: Optional[Dict[str, Any]] = None,
        system_instruction: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Generate content using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )

        response = self._get_client().post(
            f"v1beta/{model}:generateContent",
            json=payload,
        )
        response.raise_for_status()
        return response.json()

    async def generate_content_async(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Optional[Dict[str, Any]] = None,
        safety_settings: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_config: Optional[Dict[str, Any]] = None,
        system_instruction: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Generate content asynchronously using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )
        client = await self._get_async_client()
        response = await client.post(
            f"v1beta/{model}:generateContent",
            json=payload,
        )
        response.raise_for_status()
        return response.json()

    def stream_generate_content(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Optional[Dict[str, Any]] = None,
        safety_settings: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_config: Optional[Dict[str, Any]] = None,
        system_instruction: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
    ) -> Iterator[Dict[str, Any]]:
        """Stream content generation using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )


        with self._get_client().stream(
            "POST",
            f"v1beta/{model}:streamGenerateContent",
            json=payload,
        ) as response:
            response.raise_for_status()
            for line in response.iter_lines():
                if line:
                    yield json.loads(line)

    async def stream_generate_content_async(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Optional[Dict[str, Any]] = None,
        safety_settings: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_config: Optional[Dict[str, Any]] = None,
        system_instruction: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream content generation asynchronously using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )

        client = await self._get_async_client()
        async with client.stream(
            "POST",
            f"v1beta/{model}:streamGenerateContent",
            json=payload,
        ) as response:
            response.raise_for_status()
            async for line in response.aiter_lines():
                if line:
                    yield json.loads(line)

    def count_tokens(
        self,
        model: str,
        contents: List[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Count tokens in the input using the Gemini API."""
        payload = {"contents": self._convert_to_dict(contents)}
        response = self._get_client().post(
            f"v1beta/{model}:countTokens",
            json=payload,
        )
        response.raise_for_status()
        return response.json()


class ChatGoogleGenerativeAI(BaseChatModel):
    """`Google AI` chat models integration."""

    client: Optional[GeminiRestClient] = Field(default=None, exclude=True)
    async_client: Optional[GeminiRestClient] = Field(default=None, exclude=True)
    default_metadata: Sequence[Tuple[str, str]] = Field(default_factory=list)
    convert_system_message_to_human: bool = False
    cached_content: Optional[str] = None
    model: str = Field(default="gemini-pro")
    google_api_key: Optional[SecretStr] = Field(default=None)
    temperature: Optional[float] = Field(default=None)
    top_p: Optional[float] = Field(default=None)
    top_k: Optional[int] = Field(default=None)
    n: int = Field(default=1)
    max_output_tokens: Optional[int] = Field(default=None)
    safety_settings: Optional[Dict[str, str]] = Field(default=None)
    response_modalities: Optional[List[str]] = Field(default=None)
    client_options: Optional[Dict[str, Any]] = Field(default=None)
    additional_headers: Optional[Dict[str, str]] = Field(default=None)
    transport: Optional[str] = Field(default="rest")

    model_config = ConfigDict(
        populate_by_name=True,
    )

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"google_api_key": "GOOGLE_API_KEY"}

    @property
    def _llm_type(self) -> str:
        return "chat-google-generative-ai"

    @classmethod
    def is_lc_serializable(self) -> bool:
        return True

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validates params and initializes the REST client."""
        if self.temperature is not None and not 0 <= self.temperature <= 2.0:
            raise ValueError("temperature must be in the range [0.0, 2.0]")

        if self.top_p is not None and not 0 <= self.top_p <= 1:
            raise ValueError("top_p must be in the range [0.0, 1.0]")

        if self.top_k is not None and self.top_k <= 0:
            raise ValueError("top_k must be positive")

        additional_headers = self.additional_headers or {}
        self.default_metadata = tuple(additional_headers.items())

        google_api_key = None
        if isinstance(self.google_api_key, SecretStr):
            google_api_key = self.google_api_key.get_secret_value()
        else:
            google_api_key = self.google_api_key

        base_url = self.client_options.get("api_endpoint", "https://generativelanguage.googleapis.com")
        self.client = GeminiRestClient(
            api_key=google_api_key,
            base_url=base_url,
            headers=additional_headers,
        )
        self.async_client = GeminiRestClient(
            api_key=google_api_key,
            base_url=base_url,
            headers=additional_headers,
        )
        return self

    def _get_ls_params(
        self, stop: Optional[List[str]] = None, **kwargs: Any
    ) -> LangSmithParams:
        """Get standard params for tracing."""
        params = self._get_invocation_params(stop=stop, **kwargs)
        ls_params = LangSmithParams(
            ls_provider="google_genai",
            ls_model_name=self.model,
            ls_model_type="chat",
            ls_temperature=params.get("temperature", self.temperature),
        )
        if ls_max_tokens := params.get("max_output_tokens", self.max_output_tokens):
            ls_params["ls_max_tokens"] = ls_max_tokens
        if ls_stop := stop or params.get("stop", None):
            ls_params["ls_stop"] = ls_stop
        return ls_params

    def _prepare_params(
        self,
        stop: Optional[List[str]],
        generation_config: Optional[Dict[str, Any]] = None,
    ) -> GenerationConfig:
        gen_config = {
            k: v
            for k, v in {
                "candidate_count": self.n,
                "temperature": self.temperature,
                "stop_sequences": stop,
                "max_output_tokens": self.max_output_tokens,
                "top_k": self.top_k,
                "top_p": self.top_p,
                "response_modalities": self.response_modalities,
            }.items()
            if v is not None
        }
        if generation_config:
            gen_config = {**gen_config, **generation_config}
        return GenerationConfig(**gen_config)

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        *,
        tools: Optional[Sequence[Union[_ToolDict, GoogleTool]]] = None,
        functions: Optional[Sequence[_FunctionDeclarationType]] = None,
        safety_settings: Optional[SafetySettingDict] = None,
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
        tool_choice: Optional[Union[_ToolChoiceType, bool]] = None,
        **kwargs: Any,
    ) -> ChatResult:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        response = _chat_with_retry(
            request=request,
            **kwargs,
            generation_method=self.client.generate_content,
            metadata=self.default_metadata,
        )
        return _response_to_result(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        *,
        tools: Optional[Sequence[Union[_ToolDict, GoogleTool]]] = None,
        functions: Optional[Sequence[_FunctionDeclarationType]] = None,
        safety_settings: Optional[SafetySettingDict] = None,
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
        tool_choice: Optional[Union[_ToolChoiceType, bool]] = None,
        **kwargs: Any,
    ) -> ChatResult:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        response = await _achat_with_retry(
            request=request,
            **kwargs,
            generation_method=self.async_client.generate_content_async,
            metadata=self.default_metadata,
        )
        return _response_to_result(response)

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        *,
        tools: Optional[Sequence[Union[_ToolDict, GoogleTool]]] = None,
        functions: Optional[Sequence[_FunctionDeclarationType]] = None,
        safety_settings: Optional[SafetySettingDict] = None,
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
        tool_choice: Optional[Union[_ToolChoiceType, bool]] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        response = _chat_with_retry(
            request=request,
            generation_method=self.client.stream_generate_content,
            **kwargs,
            metadata=self.default_metadata,
        )

        prev_usage_metadata: UsageMetadata | None = None
        for chunk in response:
            _chat_result = _response_to_result(
                chunk, stream=True, prev_usage=prev_usage_metadata
            )
            gen = cast(ChatGenerationChunk, _chat_result.generations[0])
            message = cast(AIMessageChunk, gen.message)

            curr_usage_metadata: UsageMetadata | dict[str, int] = (
                message.usage_metadata or {}
            )

            prev_usage_metadata = (
                message.usage_metadata
                if prev_usage_metadata is None
                else UsageMetadata(
                    input_tokens=prev_usage_metadata.get("input_tokens", 0)
                    + curr_usage_metadata.get("input_tokens", 0),
                    output_tokens=prev_usage_metadata.get("output_tokens", 0)
                    + curr_usage_metadata.get("output_tokens", 0),
                    total_tokens=prev_usage_metadata.get("total_tokens", 0)
                    + curr_usage_metadata.get("total_tokens", 0),
                )
            )

            if run_manager:
                run_manager.on_llm_new_token(gen.text)
            yield gen

    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        *,
        tools: Optional[Sequence[Union[_ToolDict, GoogleTool]]] = None,
        functions: Optional[Sequence[_FunctionDeclarationType]] = None,
        safety_settings: Optional[SafetySettingDict] = None,
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
        tool_choice: Optional[Union[_ToolChoiceType, bool]] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        prev_usage_metadata: UsageMetadata | None = None
        async for chunk in await _achat_with_retry(
            request=request,
            generation_method=self.async_client.stream_generate_content_async,
            **kwargs,
            metadata=self.default_metadata,
        ):
            _chat_result = _response_to_result(
                chunk, stream=True, prev_usage=prev_usage_metadata
            )
            gen = cast(ChatGenerationChunk, _chat_result.generations[0])
            message = cast(AIMessageChunk, gen.message)

            curr_usage_metadata: UsageMetadata | dict[str, int] = (
                message.usage_metadata or {}
            )

            prev_usage_metadata = (
                message.usage_metadata
                if prev_usage_metadata is None
                else UsageMetadata(
                    input_tokens=prev_usage_metadata.get("input_tokens", 0)
                    + curr_usage_metadata.get("input_tokens", 0),
                    output_tokens=prev_usage_metadata.get("output_tokens", 0)
                    + curr_usage_metadata.get("output_tokens", 0),
                    total_tokens=prev_usage_metadata.get("total_tokens", 0)
                    + curr_usage_metadata.get("total_tokens", 0),
                )
            )

            if run_manager:
                await run_manager.on_llm_new_token(gen.text)
            yield gen

    def _prepare_request(
        self,
        messages: List[BaseMessage],
        *,
        stop: Optional[List[str]] = None,
        tools: Optional[Sequence[Union[_ToolDict, GoogleTool]]] = None,
        functions: Optional[Sequence[_FunctionDeclarationType]] = None,
        safety_settings: Optional[SafetySettingDict] = None,
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        tool_choice: Optional[Union[_ToolChoiceType, bool]] = None,
        generation_config: Optional[Dict[str, Any]] = None,
        cached_content: Optional[str] = None,
    ) -> Dict[str, Any]:
        if tool_choice and tool_config:
            raise ValueError(
                "Must specify at most one of tool_choice and tool_config, received "
                f"both:\n\n{tool_choice=}\n\n{tool_config=}"
            )
        formatted_tools = None
        if tools:
            formatted_tools = [convert_to_genai_function_declarations(tools)]
        elif functions:
            formatted_tools = [convert_to_genai_function_declarations(functions)]

        filtered_messages = []
        for message in messages:
            if isinstance(message, HumanMessage) and not message.content:
                warnings.warn(
                    "HumanMessage with empty content was removed to prevent API error"
                )
            else:
                filtered_messages.append(message)
        messages = filtered_messages

        system_instruction, history = _parse_chat_history(
            messages,
            convert_system_message_to_human=self.convert_system_message_to_human,
        )
        if tool_choice:
            if not formatted_tools:
                msg = (
                    f"Received {tool_choice=} but no {tools=}. 'tool_choice' can only "
                    f"be specified if 'tools' is specified."
                )
                raise ValueError(msg)
            all_names = [
                f.name for t in formatted_tools for f in t.function_declarations
            ]
            tool_config = _tool_choice_to_tool_config(tool_choice, all_names)

        formatted_tool_config = None
        if tool_config:
            formatted_tool_config = ToolConfig(
                function_calling_config=tool_config["function_calling_config"]
            )
        formatted_safety_settings = []
        if safety_settings:
            formatted_safety_settings = [
                SafetySetting(category=c, threshold=t)
                for c, t in safety_settings.items()
            ]

        # Construct the full model path
        model_path = f"models/{self.model}"

        request = {
            "model": model_path,
            "contents": history,
            "tools": formatted_tools,
            "tool_config": formatted_tool_config,
            "safety_settings": formatted_safety_settings,
            "generation_config": self._prepare_params(
                stop, generation_config=generation_config
            ),
            "cached_content": cached_content,
        }
        if system_instruction:
            request["system_instruction"] = system_instruction

        return request

    def get_num_tokens(self, text: str) -> int:
        """Get the number of tokens present in the text.

        Useful for checking if an input will fit in a model's context window.

        Args:
            text: The string input to tokenize.

        Returns:
            The integer number of tokens in the text.
        """
        result = self.client.count_tokens(
            model=self.model,
            contents=[{"parts": [{"text": text}]}],
        )
        return result["total_tokens"]

    def with_structured_output(
        self,
        schema: Union[Dict, Type[BaseModel]],
        *,
        include_raw: bool = False,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, Union[Dict, BaseModel]]:
        _ = kwargs.pop("method", None)
        _ = kwargs.pop("strict", None)
        if kwargs:
            raise ValueError(f"Received unsupported arguments {kwargs}")
        tool_name = _get_tool_name(schema)  # type: ignore[arg-type]
        if isinstance(schema, type) and is_basemodel_subclass_safe(schema):
            parser: OutputParserLike = PydanticToolsParser(
                tools=[schema], first_tool_only=True
            )
        else:
            global WARNED_STRUCTURED_OUTPUT_JSON_MODE
            warnings.warn(
                "ChatGoogleGenerativeAI.with_structured_output with dict schema has "
                "changed recently to align with behavior of other LangChain chat "
                "models. More context: "
                "https://github.com/langchain-ai/langchain-google/pull/772"
            )
            WARNED_STRUCTURED_OUTPUT_JSON_MODE = True
            parser = JsonOutputKeyToolsParser(key_name=tool_name, first_tool_only=True)
        tool_choice = tool_name if self._supports_tool_choice else None
        try:
            llm = self.bind_tools(
                [schema],
                tool_choice=tool_choice,
                ls_structured_output_format={
                    "kwargs": {"method": "function_calling"},
                    "schema": convert_to_openai_tool(schema),
                },
            )
        except Exception:
            llm = self.bind_tools([schema], tool_choice=tool_choice)
        if include_raw:
            parser_with_fallback = RunnablePassthrough.assign(
                parsed=itemgetter("raw") | parser, parsing_error=lambda _: None
            ).with_fallbacks(
                [RunnablePassthrough.assign(parsed=lambda _: None)],
                exception_key="parsing_error",
            )
            return {"raw": llm} | parser_with_fallback
        else:
            return llm | parser

    def bind_tools(
        self,
        tools: Sequence[
            dict[str, Any] | type | Callable[..., Any] | BaseTool | GoogleTool
        ],
        tool_config: Optional[Union[Dict, _ToolConfigDict]] = None,
        *,
        tool_choice: Optional[Union[_ToolChoiceType, bool]] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        """Bind tool-like objects to this chat model.

        Assumes model is compatible with google-generativeAI tool-calling API.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be a pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            **kwargs: Any additional parameters to pass to the
                :class:`~langchain.runnable.Runnable` constructor.
        """
        if tool_choice and tool_config:
            raise ValueError(
                "Must specify at most one of tool_choice and tool_config, received "
                f"both:\n\n{tool_choice=}\n\n{tool_config=}"
            )
        try:
            formatted_tools: list = [convert_to_openai_tool(tool) for tool in tools]  # type: ignore[arg-type]
        except Exception:
            formatted_tools = [
                tool_to_dict(convert_to_genai_function_declarations(tools))
            ]
        if tool_choice:
            kwargs["tool_choice"] = tool_choice
        elif tool_config:
            kwargs["tool_config"] = tool_config
        else:
            pass
        return self.bind(tools=formatted_tools, **kwargs)

    @property
    def _supports_tool_choice(self) -> bool:
        return (
            "gemini-1.5-pro" in self.model
            or "gemini-1.5-flash" in self.model
            or "gemini-2" in self.model
        )


def _get_tool_name(
    tool: Union[_ToolDict, GoogleTool, Dict],
) -> str:
    try:
        genai_tool = tool_to_dict(convert_to_genai_function_declarations([tool]))
        return [f["name"] for f in genai_tool["function_declarations"]][0]  # type: ignore[index]
    except ValueError as e:  # other TypedDict
        if is_typeddict(tool):
            return convert_to_openai_tool(cast(Dict, tool))["function"]["name"]
        else:
            raise e

def _tool_choice_to_tool_config(
    tool_choice: Union[str, bool, Dict[str, Any]], all_names: List[str]
) -> Dict[str, Any]:
    """Convert tool_choice to tool_config format."""
    if isinstance(tool_choice, bool):
        return {
            "function_calling_config": {
                "mode": "AUTO" if tool_choice else "NONE",
            }
        }
    elif isinstance(tool_choice, str):
        if tool_choice not in all_names:
            raise ValueError(
                f"Tool choice {tool_choice} not found in available tools: {all_names}"
            )
        return {
            "function_calling_config": {
                "mode": "ANY",
                "allowed_function_names": [tool_choice],
            }
        }
    else:
        return tool_choice

def convert_to_genai_function_declarations(
    tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable[..., Any], BaseTool]]
) -> Dict[str, Any]:
    """Convert tools to Gemini function declarations format."""
    function_declarations = []
    for tool in tools:
        if isinstance(tool, dict):
            fn = tool.get("function", {})
            fn_parameters = fn.get("parameters", {})
            function_declarations.append({
                "name": fn.get("name", ""),
                "description": fn.get("description", ""),
                "parameters": {
                    "type": "object",
                    "properties": fn_parameters.get("properties", {}),
                    "required": fn_parameters.get("required", []),
                },
            })
        elif isinstance(tool, type) and issubclass(tool, BaseModel):
            schema = tool.model_json_schema()
            function_declarations.append({
                "name": schema.get("title", ""),
                "description": schema.get("description", ""),
                "parameters": {
                    "type": "object",
                    "properties": {
                        name: {
                            "type": prop.get("type", "string"),
                            "description": prop.get("description", ""),
                        }
                        for name, prop in schema.get("properties", {}).items()
                    },
                    "required": schema.get("required", []),
                },
            })
        elif callable(tool):
            # For callables, we'll create a basic function declaration
            function_declarations.append({
                "name": tool.__name__,
                "description": tool.__doc__ or "",
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            })
        elif isinstance(tool, BaseTool):
            function_declarations.append({
                "name": tool.name,
                "description": tool.description,
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            })
    return {"functionDeclarations": function_declarations}

def is_basemodel_subclass_safe(cls: Type[Any]) -> bool:
    """Check if a class is a safe subclass of BaseModel."""
    try:
        return issubclass(cls, BaseModel)
    except TypeError:
        return False

def tool_to_dict(tool: Dict[str, Any]) -> Dict[str, Any]:
    """Convert a tool to a dictionary format."""
    return tool

def image_bytes_to_b64_string(image_bytes: bytes, image_format: str = "jpeg") -> str:
    """Convert image bytes to base64 string."""
    import base64
    return f"data:image/{image_format};base64,{base64.b64encode(image_bytes).decode('utf-8')}"



import os
from logging import getLogger
from pathlib import Path
from typing import Optional

import yaml
from httpx import Auth

from .apikey import ApiKey
from .clientcredentials import ClientCredentials
from .devicemode import DeviceMode
from .types import BlaxelAuth, CredentialsType

logger = getLogger(__name__)

def get_credentials() -> Optional[CredentialsType]:
    """
    Get credentials from environment variables or config file.

    Returns:
        Optional[CredentialsType]: The credentials or None if not found
    """
    if os.environ.get("BL_API_KEY"):
        return CredentialsType(
            api_key=os.environ.get("BL_API_KEY"),
            workspace=os.environ.get("BL_WORKSPACE")
        )

    if os.environ.get("BL_CLIENT_CREDENTIALS"):
        return CredentialsType(
            client_credentials=os.environ.get("BL_CLIENT_CREDENTIALS"),
            workspace=os.environ.get("BL_WORKSPACE")
        )

    try:
        home_dir = Path.home()
        config_path = home_dir / '.blaxel' / 'config.yaml'

        with open(config_path, encoding='utf-8') as f:
            config_json = yaml.safe_load(f)

        workspace_name = os.environ.get("BL_WORKSPACE") or config_json.get("context", {}).get("workspace")

        for workspace in config_json.get("workspaces", []):
            if workspace.get("name") == workspace_name:
                credentials = workspace.get("credentials", {})
                credentials["workspace"] = workspace_name
                return CredentialsType(
                    workspace=credentials["workspace"],
                    api_key=credentials.get("api_key"),
                    client_credentials=credentials.get("client_credentials"),
                    refresh_token=credentials.get("refresh_token"),
                    access_token=credentials.get("access_token"),
                    device_code=credentials.get("device_code"),
                    expires_in=credentials.get("expires_in")
                )

        return None
    except Exception:
        return None


def auth(env: str, base_url: str) -> BlaxelAuth:
    """
    Create and return the appropriate credentials object based on available credentials.

    Returns:
        Credentials: The credentials object
    """
    credentials = get_credentials()

    if not credentials:
        return None

    if credentials.api_key:
        logger.debug(f"Using API key for authentication")
        return ApiKey(credentials, credentials.workspace, base_url)

    if credentials.client_credentials:
        logger.debug(f"Using client credentials for authentication")
        return ClientCredentials(credentials, credentials.workspace, base_url)

    if credentials.device_code:
        logger.debug(f"Using device code for authentication")
        return DeviceMode(credentials, credentials.workspace, base_url)

    return BlaxelAuth(credentials, credentials.workspace, base_url)

__all__ = ["BlaxelAuth", "CredentialsType"]


import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor

from crewai.tools import BaseTool

from .common import create_model_from_json_schema
from .types import Tool


def get_crewai_tools(tools: list[Tool]) -> list[BaseTool]:
    class CrewAITool(BaseTool):
        _tool: Tool

        def __init__(self, tool: Tool):
            super().__init__(
                name=tool.name,
                description=tool.description,
                args_schema=create_model_from_json_schema(tool.input_schema),
            )
            self._tool = tool

        def _run(self, *args, **kwargs):
            return self._tool.sync_coroutine(**kwargs)

    return [CrewAITool(tool) for tool in tools]


from typing import Any

from langchain_core.tools import StructuredTool
from mcp.types import (CallToolResult, EmbeddedResource, ImageContent,
                       TextContent)

from .types import Tool, ToolException

NonTextContent = ImageContent | EmbeddedResource

def get_langchain_tool(tool: Tool) -> StructuredTool:
    async def langchain_coroutine(
       **arguments: dict[str, Any],
    ) -> tuple[str | list[str], list[NonTextContent] | None]:
        result: CallToolResult = await tool.coroutine(**arguments)
        text_contents: list[TextContent] = []
        non_text_contents = []
        for content in result.content:
            if isinstance(content, TextContent):
                text_contents.append(content)
            else:
                non_text_contents.append(content)

        tool_content: str | list[str] = [content.text for content in text_contents]
        if len(text_contents) == 1:
            tool_content = tool_content[0]

        if result.isError:
            raise ToolException(tool_content)

        return tool_content, non_text_contents or None

    return StructuredTool(
        name=tool.name,
        description=tool.description,
        args_schema=tool.input_schema,
        coroutine=langchain_coroutine,
        sync_coroutine=tool.sync_coroutine,
    )

def get_langchain_tools(tools: list[Tool]) -> list[StructuredTool]:
    return [get_langchain_tool(tool) for tool in tools]


import time
import traceback
from contextlib import asynccontextmanager
from logging import getLogger

import uvicorn
from asgi_correlation_id import CorrelationIdMiddleware
from fastapi import FastAPI, Request, Response, status
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from llama_index.core.agent.workflow import (AgentOutput, ReActAgent,
                                             ToolCallResult)
from llama_index.core.workflow import Context
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

from blaxel.instrumentation.span import SpanManager
from blaxel.models import bl_model
from blaxel.tools import bl_tools

logger = getLogger(__name__)

@asynccontextmanager
async def lifespan(app: FastAPI):
    logger.info("Server running on port 1338")
    yield
    logger.info("Server shutting down")

app = FastAPI(lifespan=lifespan)
app.add_middleware(CorrelationIdMiddleware)

@app.middleware("http")
async def log_requests(request: Request, call_next):
    start_time = time.time()

    response: Response = await call_next(request)

    process_time = (time.time() - start_time) * 1000
    formatted_process_time = '{0:.2f}'.format(process_time)
    rid_header = response.headers.get("X-Request-Id")
    request_id = rid_header or response.headers.get("X-Blaxel-Request-Id")
    logger.info(f"{request.method} {request.url.path} {response.status_code} {formatted_process_time}ms rid={request_id}")

    return response

@app.exception_handler(Exception)
async def validation_exception_handler(request: Request, e: Exception):
    logger.error(f"Error on request {request.method} {request.url.path}: {e}")

    # Get the full traceback information
    tb_str = traceback.format_exception(type(e), e, e.__traceback__)
    logger.error(f"Stacktrace: {''.join(tb_str)}")
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content=jsonable_encoder({"detail": str(e)}),
    )

@app.post("/")
async def handle_request(request: Request):
    start = time.time()
    tools = await bl_tools(["blaxel-search"]).to_llamaindex()
    logger.info(f"Loaded tools in {round(time.time() - start, 4)} seconds")

    start = time.time()
    model = await bl_model("gpt-4o-mini").to_llamaindex()
    logger.info(f"Retrieved model in {round(time.time() - start, 4)} seconds")

    with SpanManager("blaxel-llamaindex").create_active_span("agent-request", {}) as span:
        body = await request.json()
        agent = ReActAgent(
            llm=model,
            tools=tools,
            system_prompt="You are a helpful assistant. Maximum number of tool call is 1."
        )
        context = Context(agent)
        handler = agent.run(body.get("inputs", ""), ctx=context)
        async for ev in handler.stream_events():
            if isinstance(ev, ToolCallResult):
                logger.info(f"Call {ev.tool_name} with {ev.tool_kwargs}")
        response: AgentOutput = await handler
        span.set_attribute("agent.result", response.response.blocks[-1].text)
        return Response(
            response.response.blocks[-1].text,
            media_type="text/plain"
        )

FastAPIInstrumentor.instrument_app(app)

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=1338, log_level="critical", loop="asyncio")



import asyncio
import json
from contextlib import AsyncExitStack
from logging import getLogger
from typing import Any, cast

from mcp import ClientSession
from mcp.types import CallToolResult
from mcp.types import Tool as MCPTool

from ..cache import find_from_cache
from ..client import client
from ..client.api.functions import get_function
from ..client.models import Function
from ..common.env import env
from ..common.settings import settings
from ..instrumentation.span import SpanManager
from ..mcp.client import websocket_client
from .types import Tool

logger = getLogger(__name__)


def convert_mcp_tool_to_blaxel_tool(
    url: str,
    tool: MCPTool,
) -> Tool:
    """Convert an MCP tool to a LangChain tool.

    NOTE: this tool can be executed only in a context of an active MCP client session.

    Args:
        session: MCP client session
        tool: MCP tool to convert

    Returns:
        a LangChain tool
    """

    async def initialize_and_call_tool(
        *args: Any,
        **arguments: dict[str, Any],
    ) -> CallToolResult:
        with SpanManager("blaxel-tracer").create_active_span(tool.name, {"tool.name": tool.name, "tool.args": json.dumps(arguments)}):
            exit_stack = AsyncExitStack()
            read, write = await exit_stack.enter_async_context(websocket_client(url, settings.headers))
            session = cast(ClientSession, await exit_stack.enter_async_context(ClientSession(read, write)))
            await session.initialize()
            logger.debug(f"Calling tool {tool.name} with arguments {arguments}")
            call_tool_result = await session.call_tool(tool.name, arguments)
            logger.debug(f"Tool {tool.name} returned {call_tool_result}")
            await exit_stack.aclose()
            return call_tool_result

    async def call_tool(
        *args: Any,
        **arguments: dict[str, Any],
    ) -> CallToolResult:
        return await initialize_and_call_tool(*args, **arguments)

    def sync_call_tool(*args: Any, **arguments: dict[str, Any]) -> CallToolResult:
        try:
            loop = asyncio.get_running_loop()
            return loop.run_until_complete(initialize_and_call_tool(*args, **arguments))
        except RuntimeError:
            return asyncio.run(initialize_and_call_tool(*args, **arguments))

    return Tool(
        name=tool.name,
        description=tool.description or "",
        input_schema=tool.inputSchema,
        coroutine=call_tool,
        sync_coroutine=sync_call_tool,
        response_format="content_and_artifact",
    )


async def load_mcp_tools(url: str) -> list[Tool]:
    """Load all available MCP tools and convert them to LangChain tools."""
    exit_stack = AsyncExitStack()
    read, write = await exit_stack.enter_async_context(websocket_client(url, settings.headers))
    session = cast(ClientSession, await exit_stack.enter_async_context(ClientSession(read, write)))
    await session.initialize()

    tools = await session.list_tools()
    await exit_stack.aclose()
    return [convert_mcp_tool_to_blaxel_tool(url, tool) for tool in tools.tools]

class BlTools:
    server_name_to_tools: dict[str, list[Tool]] = {}

    def __init__(self, functions: list[str]):
        self.exit_stack = AsyncExitStack()
        self.sessions: dict[str, ClientSession] = {}
        self.functions = functions

    def _external_url(self, name: str) -> str:
        return f"{settings.run_url}/{settings.auth.workspace_name}/functions/{name}"

    def _url(self, name: str) -> str:
        env_var = name.replace("-", "_").upper()
        if env[f"BL_FUNCTION_{env_var}_URL"]:
            return env[f"BL_FUNCTION_{env_var}_URL"]
        elif env[f"BL_FUNCTION_{env_var}_SERVICE_NAME"]:
            return f"https://{env[f'BL_FUNCTION_{env_var}_SERVICE_NAME']}.{settings.run_internal_hostname}"
        return self._external_url(name)

    def _fallback_url(self, name: str) -> str | None:
        if self._external_url(name) != self._url(name):
            return self._external_url(name)
        return None

    def get_tools(self) -> list[Tool]:
        """Get a list of all tools from all connected servers."""
        all_tools: list[Tool] = []
        for server_tools in BlTools.server_name_to_tools.values():
            all_tools.extend(server_tools)
        return all_tools

    async def to_langchain(self):
        from .langchain import get_langchain_tools

        await self.intialize()
        return get_langchain_tools(self.get_tools())

    async def to_llamaindex(self):
        from .llamaindex import get_llamaindex_tools

        await self.intialize()
        return get_llamaindex_tools(self.get_tools())

    async def to_crewai(self):
        from .crewai import get_crewai_tools

        await self.intialize()
        return get_crewai_tools(self.get_tools())

    async def to_openai(self):
        from .openai import get_openai_tools

        await self.intialize()
        return get_openai_tools(self.get_tools())

    async def connect_to_server_via_websocket(self, name: str):
        # Create and store the connection
        try:
            url = self._url(name)
            await self._initialize_session_and_load_tools(name, url)
        except Exception as e:
            if not self._fallback_url(name):
                raise e
            url = self._fallback_url(name)
            await self._initialize_session_and_load_tools(name, url)

    async def _initialize_session_and_load_tools(
        self, name: str, url: str
    ) -> None:
        """Initialize a session and load tools from it.

        Args:
            name: Name to identify this server connection
            url: The URL to connect to
        """
        logger.debug(f"Initializing session and loading tools from {url}")
        server_tools = await load_mcp_tools(url)
        logger.debug(f"Loaded {len(server_tools)} tools from {url}")
        BlTools.server_name_to_tools[name] = server_tools

    async def _get_function(self, tool) -> Function | None:
        cache_data = await find_from_cache('Function', tool)
        if cache_data:
            return Function.from_dict(cache_data)
        try:
            return await get_function.asyncio(client=client, function_name=tool)
        except Exception as e:
            return None

    async def intialize(self) -> "BlTools":
        try:
            functions = []
            for name in self.functions:
                if name in BlTools.server_name_to_tools:
                    continue
                function = await self._get_function(name)
                if function:
                    functions.append(function)
                else:
                    if not env[f"BL_FUNCTION_{name.replace('-', '_').upper()}_URL"]:
                        logger.warning(f"Function {name} not loaded, skipping")
                try:
                    await self.connect_to_server_via_websocket(name)
                except Exception as e:
                    logger.warning(f"Failed to connect to server {name}: {e}")
            return self
        except Exception:
            await self.exit_stack.aclose()
            raise


def bl_tools(functions: list[str]) -> BlTools:
    return BlTools(functions)



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.token_rate_metric import TokenRateMetric


T = TypeVar("T", bound="TokenRateMetrics")


@_attrs_define
class TokenRateMetrics:
    """Token rate metrics

    Attributes:
        token_rate (Union[Unset, TokenRateMetric]): Token rate metric
        token_rate_input (Union[Unset, TokenRateMetric]): Token rate metric
        token_rate_output (Union[Unset, TokenRateMetric]): Token rate metric
    """

    token_rate: Union[Unset, "TokenRateMetric"] = UNSET
    token_rate_input: Union[Unset, "TokenRateMetric"] = UNSET
    token_rate_output: Union[Unset, "TokenRateMetric"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        token_rate: Union[Unset, dict[str, Any]] = UNSET
        if self.token_rate and not isinstance(self.token_rate, Unset) and not isinstance(self.token_rate, dict):
            token_rate = self.token_rate.to_dict()
        elif self.token_rate and isinstance(self.token_rate, dict):
            token_rate = self.token_rate

        token_rate_input: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.token_rate_input
            and not isinstance(self.token_rate_input, Unset)
            and not isinstance(self.token_rate_input, dict)
        ):
            token_rate_input = self.token_rate_input.to_dict()
        elif self.token_rate_input and isinstance(self.token_rate_input, dict):
            token_rate_input = self.token_rate_input

        token_rate_output: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.token_rate_output
            and not isinstance(self.token_rate_output, Unset)
            and not isinstance(self.token_rate_output, dict)
        ):
            token_rate_output = self.token_rate_output.to_dict()
        elif self.token_rate_output and isinstance(self.token_rate_output, dict):
            token_rate_output = self.token_rate_output

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if token_rate is not UNSET:
            field_dict["tokenRate"] = token_rate
        if token_rate_input is not UNSET:
            field_dict["tokenRateInput"] = token_rate_input
        if token_rate_output is not UNSET:
            field_dict["tokenRateOutput"] = token_rate_output

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.token_rate_metric import TokenRateMetric

        if not src_dict:
            return None
        d = src_dict.copy()
        _token_rate = d.pop("tokenRate", UNSET)
        token_rate: Union[Unset, TokenRateMetric]
        if isinstance(_token_rate, Unset):
            token_rate = UNSET
        else:
            token_rate = TokenRateMetric.from_dict(_token_rate)

        _token_rate_input = d.pop("tokenRateInput", UNSET)
        token_rate_input: Union[Unset, TokenRateMetric]
        if isinstance(_token_rate_input, Unset):
            token_rate_input = UNSET
        else:
            token_rate_input = TokenRateMetric.from_dict(_token_rate_input)

        _token_rate_output = d.pop("tokenRateOutput", UNSET)
        token_rate_output: Union[Unset, TokenRateMetric]
        if isinstance(_token_rate_output, Unset):
            token_rate_output = UNSET
        else:
            token_rate_output = TokenRateMetric.from_dict(_token_rate_output)

        token_rate_metrics = cls(
            token_rate=token_rate,
            token_rate_input=token_rate_input,
            token_rate_output=token_rate_output,
        )

        token_rate_metrics.additional_properties = d
        return token_rate_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.runtime_startup_probe import RuntimeStartupProbe


T = TypeVar("T", bound="Runtime")


@_attrs_define
class Runtime:
    """Set of configurations for a deployment

    Attributes:
        args (Union[Unset, list[Any]]): The arguments to pass to the deployment runtime
        command (Union[Unset, list[Any]]): The command to run the deployment
        cpu (Union[Unset, int]): The CPU for the deployment in cores, only available for private cluster
        endpoint_name (Union[Unset, str]): Endpoint Name of the model. In case of hf_private_endpoint, it is the
            endpoint name. In case of hf_public_endpoint, it is not used.
        envs (Union[Unset, list[Any]]): The env variables to set in the deployment. Should be a list of Kubernetes
            EnvVar types
        image (Union[Unset, str]): The Docker image for the deployment
        memory (Union[Unset, int]): The memory for the deployment in MB
        metric_port (Union[Unset, int]): The port to serve the metrics on
        model (Union[Unset, str]): The slug name of the origin model at HuggingFace.
        organization (Union[Unset, str]): The organization of the model
        serving_port (Union[Unset, int]): The port to serve the model on
        startup_probe (Union[Unset, RuntimeStartupProbe]): The readiness probe. Should be a Kubernetes Probe type
        type_ (Union[Unset, str]): The type of origin for the deployment (hf_private_endpoint, hf_public_endpoint)
    """

    args: Union[Unset, list[Any]] = UNSET
    command: Union[Unset, list[Any]] = UNSET
    cpu: Union[Unset, int] = UNSET
    endpoint_name: Union[Unset, str] = UNSET
    envs: Union[Unset, list[Any]] = UNSET
    image: Union[Unset, str] = UNSET
    memory: Union[Unset, int] = UNSET
    metric_port: Union[Unset, int] = UNSET
    model: Union[Unset, str] = UNSET
    organization: Union[Unset, str] = UNSET
    serving_port: Union[Unset, int] = UNSET
    startup_probe: Union[Unset, "RuntimeStartupProbe"] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        args: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.args, Unset):
            args = self.args

        command: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.command, Unset):
            command = self.command

        cpu = self.cpu

        endpoint_name = self.endpoint_name

        envs: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.envs, Unset):
            envs = self.envs

        image = self.image

        memory = self.memory

        metric_port = self.metric_port

        model = self.model

        organization = self.organization

        serving_port = self.serving_port

        startup_probe: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.startup_probe
            and not isinstance(self.startup_probe, Unset)
            and not isinstance(self.startup_probe, dict)
        ):
            startup_probe = self.startup_probe.to_dict()
        elif self.startup_probe and isinstance(self.startup_probe, dict):
            startup_probe = self.startup_probe

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if args is not UNSET:
            field_dict["args"] = args
        if command is not UNSET:
            field_dict["command"] = command
        if cpu is not UNSET:
            field_dict["cpu"] = cpu
        if endpoint_name is not UNSET:
            field_dict["endpointName"] = endpoint_name
        if envs is not UNSET:
            field_dict["envs"] = envs
        if image is not UNSET:
            field_dict["image"] = image
        if memory is not UNSET:
            field_dict["memory"] = memory
        if metric_port is not UNSET:
            field_dict["metricPort"] = metric_port
        if model is not UNSET:
            field_dict["model"] = model
        if organization is not UNSET:
            field_dict["organization"] = organization
        if serving_port is not UNSET:
            field_dict["servingPort"] = serving_port
        if startup_probe is not UNSET:
            field_dict["startupProbe"] = startup_probe
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.runtime_startup_probe import RuntimeStartupProbe

        if not src_dict:
            return None
        d = src_dict.copy()
        args = cast(list[Any], d.pop("args", UNSET))

        command = cast(list[Any], d.pop("command", UNSET))

        cpu = d.pop("cpu", UNSET)

        endpoint_name = d.pop("endpointName", UNSET)

        envs = cast(list[Any], d.pop("envs", UNSET))

        image = d.pop("image", UNSET)

        memory = d.pop("memory", UNSET)

        metric_port = d.pop("metricPort", UNSET)

        model = d.pop("model", UNSET)

        organization = d.pop("organization", UNSET)

        serving_port = d.pop("servingPort", UNSET)

        _startup_probe = d.pop("startupProbe", UNSET)
        startup_probe: Union[Unset, RuntimeStartupProbe]
        if isinstance(_startup_probe, Unset):
            startup_probe = UNSET
        else:
            startup_probe = RuntimeStartupProbe.from_dict(_startup_probe)

        type_ = d.pop("type", UNSET)

        runtime = cls(
            args=args,
            command=command,
            cpu=cpu,
            endpoint_name=endpoint_name,
            envs=envs,
            image=image,
            memory=memory,
            metric_port=metric_port,
            model=model,
            organization=organization,
            serving_port=serving_port,
            startup_probe=startup_probe,
            type_=type_,
        )

        runtime.additional_properties = d
        return runtime

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
from contextlib import asynccontextmanager
from typing import Any
from urllib.parse import urljoin, urlparse

import anyio
import mcp.types as types
from anyio.abc import TaskStatus
from anyio.streams.memory import (MemoryObjectReceiveStream,
                                  MemoryObjectSendStream)
from websockets.client import WebSocketClientProtocol
from websockets.client import connect as ws_connect

logger = logging.getLogger(__name__)


def remove_request_params(url: str) -> str:
    return urljoin(url, urlparse(url).path)


@asynccontextmanager
async def websocket_client(
    url: str,
    headers: dict[str, Any] | None = None,
    timeout: float = 30,
):
    """
    Client transport for WebSocket.

    The `timeout` parameter controls connection timeout.
    """
    read_stream: MemoryObjectReceiveStream[types.JSONRPCMessage | Exception]
    read_stream_writer: MemoryObjectSendStream[types.JSONRPCMessage | Exception]

    write_stream: MemoryObjectSendStream[types.JSONRPCMessage]
    write_stream_reader: MemoryObjectReceiveStream[types.JSONRPCMessage]

    read_stream_writer, read_stream = anyio.create_memory_object_stream(0)
    write_stream, write_stream_reader = anyio.create_memory_object_stream(0)

    async with anyio.create_task_group() as tg:
        try:
            # Convert http(s):// to ws(s)://
            ws_url = url.replace("http://", "ws://").replace("https://", "wss://")
            logger.debug(f"Connecting to WebSocket endpoint: {remove_request_params(ws_url)}")

            async with ws_connect(ws_url, extra_headers=headers, open_timeout=timeout) as websocket:
                logger.debug("WebSocket connection established")

                async def ws_reader(
                    websocket: WebSocketClientProtocol,
                    task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
                ):
                    try:
                        task_status.started()
                        async for message in websocket:
                            logger.debug(f"Received WebSocket message: {message}")
                            try:
                                parsed_message = types.JSONRPCMessage.model_validate_json(message)
                                logger.debug(f"Received server message: {parsed_message}")
                                await read_stream_writer.send(parsed_message)
                            except Exception as exc:
                                logger.error(f"Error parsing server message: {exc}")
                                await read_stream_writer.send(exc)
                    except Exception as exc:
                        logger.error(f"Error in ws_reader: {exc}")
                        await read_stream_writer.send(exc)
                    finally:
                        await read_stream_writer.aclose()

                async def ws_writer(websocket: WebSocketClientProtocol):
                    try:
                        async with write_stream_reader:
                            async for message in write_stream_reader:
                                logger.debug(f"Sending client message: {message}")
                                await websocket.send(
                                    message.model_dump_json(
                                        by_alias=True,
                                        exclude_none=True,
                                    )
                                )
                                logger.debug("Client message sent successfully")
                    except Exception as exc:
                        logger.error(f"Error in ws_writer: {exc}")
                    finally:
                        await write_stream.aclose()

                await tg.start(ws_reader, websocket)
                tg.start_soon(ws_writer, websocket)

                try:
                    yield read_stream, write_stream
                finally:
                    tg.cancel_scope.cancel()
        finally:
            await read_stream_writer.aclose()
            await write_stream.aclose()



import json
from logging import getLogger
from typing import Any, Awaitable

from ..cache import find_from_cache
from ..client import client
from ..client.api.agents import get_agent
from ..client.models import Agent
from ..common.env import env
from ..common.settings import settings
from ..instrumentation.span import SpanManager

logger = getLogger(__name__)

class BlAgent:
    def __init__(self, name: str):
        self.name = name


    @property
    def external_url(self):
        return f"{settings.run_url}/{settings.workspace}/agents/{self.name}"

    @property
    def fallback_url(self):
        if self.external_url != self.url:
            return self.external_url
        return None

    @property
    def url(self):
        env_var = self.name.replace("-", "_").upper()
        if env[f"BL_AGENT_{env_var}_URL"]:
            return env[f"BL_AGENT_{env_var}_URL"]
        if f"BL_AGENT_{env_var}_SERVICE_NAME" in settings.env:
            return f"https://{settings.env[f'BL_AGENT_{env_var}_SERVICE_NAME']}.{settings.run_internal_hostname}"
        return self.external_url

    def call(self, url, input_data):
        body = input_data
        if not isinstance(body, str):
            body = json.dumps(body)

        return client.get_httpx_client().post(
            url,
            headers={
                'Content-Type': 'application/json'
            },
            data=body
        )

    async def acall(self, input_data):
        logger.debug(f"Agent Calling: {self.name}")
        body = input_data
        if not isinstance(body, str):
            body = json.dumps(body)

        return await client.get_async_httpx_client().post(
            self.url,
            headers={
                'Content-Type': 'application/json'
            },
            data=body
        )

    def run(self, input: Any) -> str:
        with SpanManager("blaxel-tracer").create_active_span(self.name, {"agent.name": self.name, "agent.args": json.dumps(input)}):
            logger.debug(f"Agent Calling: {self.name}")
            response = self.call(self.url, input)
            if response.status_code >= 400:
                raise Exception(f"Agent {self.name} returned status code {response.status_code} with body {response.text}")
            return response.text

    async def arun(self, input: Any) -> Awaitable[str]:
        logger.debug(f"Agent Calling: {self.name}")
        response = await self.acall(input)
        if response.status_code >= 400:
            raise Exception(f"Agent {self.name} returned status code {response.status_code} with body {response.text}")
        return response.text

    def __str__(self):
        return f"Agent {self.name}"

    def __repr__(self):
        return self.__str__()

def bl_agent(name: str):
    return BlAgent(name)

async def get_agent_metadata(name):
    cache_data = await find_from_cache('Agent', name)
    if cache_data:
        return Agent.from_dict(cache_data)
    try:
        return await get_agent.asyncio(client=client, agent_name=name)
    except Exception as e:
        return None




import nest_asyncio

nest_asyncio.apply()

import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from blaxel.tools import bl_tools

logger = getLogger(__name__)

async def test_mcp_tools_langchain():
    tools = await bl_tools(["blaxel-search"]).to_langchain()
    if len(tools) == 0:
        raise Exception("No tools found")
    result = await tools[0].ainvoke({ "query": "What is the capital of France?"})
    logger.info(result)

async def test_mcp_tools_llamaindex():
    tools = await bl_tools(["blaxel-search"]).to_llamaindex()
    if len(tools) == 0:
        raise Exception("No tools found")
    result = await tools[0].acall(query="What is the capital of France?")
    logger.info(result)

async def test_mcp_tools_crewai():
    tools = await bl_tools(["blaxel-search"]).to_crewai()
    if len(tools) == 0:
        raise Exception("No tools found")
    result = tools[0].run(query="What is the capital of France?")
    logger.info(result)

async def main():
    await test_mcp_tools_langchain()
    await test_mcp_tools_llamaindex()
    await test_mcp_tools_crewai()

if __name__ == "__main__":
    asyncio.run(main())